@@ -6,7 +6,7 @@ import inspect
 import logging
 import os
 import re
-import subprocess
+import subprocess  # nosec
 import sys
 import textwrap
 import warnings
@@ -15,12 +15,14 @@ from pathlib import Path
 from pprint import pformat
 from typing import TYPE_CHECKING, Any, Callable
 
-from deprecated import deprecated
-from setuptools.dist import Distribution
-
 if TYPE_CHECKING:
+    # avoid importing 'packaging' because setuptools-git-versioning can be installed using sdist
+    # where 'packaging' is not installed yet
     from packaging.version import Version
 
+    # used only for mypy check
+    from setuptools.dist import Distribution
+
 DEFAULT_TEMPLATE = "{tag}"
 DEFAULT_DEV_TEMPLATE = "{tag}.post{ccount}+git.{sha}"
 DEFAULT_DIRTY_TEMPLATE = "{tag}.post{ccount}+git.{sha}.dirty"
@@ -28,8 +30,9 @@ DEFAULT_STARTING_VERSION = "0.0.1"
 DEFAULT_SORT_BY = "creatordate"
 ENV_VARS_REGEXP = re.compile(r"\{env:(?P<name>[^:}]+):?(?P<default>[^}]+\}*)?\}", re.IGNORECASE | re.UNICODE)
 TIMESTAMP_REGEXP = re.compile(r"\{timestamp:?(?P<fmt>[^:}]+)?\}", re.IGNORECASE | re.UNICODE)
-LOCAL_REGEXP = re.compile(r"[^a-z\d.]+", re.IGNORECASE)
-VERSION_PREFIX_REGEXP = re.compile(r"^[^\d]+", re.IGNORECASE | re.UNICODE)
+
+# https://github.com/pypa/setuptools/blob/bc39d28bda2a1faee6680ae30e42526b9d775151/setuptools/command/dist_info.py#L108-L131
+UNSUPPORTED_SYMBOL_REGEXP = re.compile(r"[^\w\d]+", re.IGNORECASE | re.UNICODE)
 
 LOG_FORMAT = "[%(asctime)s] %(levelname)+8s: %(message)s"
 # setuptools v60.2.0 changed default logging level to DEBUG: https://github.com/pypa/setuptools/pull/2974
@@ -55,6 +58,7 @@ DEFAULT_CONFIG = {
     "count_commits_from_version_file": False,
     "tag_formatter": None,
     "branch_formatter": None,
+    "tag_filter": None,
     "sort_by": DEFAULT_SORT_BY,
 }
 
@@ -62,7 +66,7 @@ log = logging.getLogger(__name__)
 
 
 def _exec(cmd: str, root: str | os.PathLike | None = None) -> list[str]:
-    log.log(DEBUG, "Executing '%s' at '%s'", cmd, root or os.getcwd())
+    log.log(DEBUG, "Executing %r at '%s'", cmd, root or os.getcwd())
     try:
         stdout = subprocess.check_output(cmd, shell=True, text=True, cwd=root)  # nosec
     except subprocess.CalledProcessError as e:
@@ -72,6 +76,7 @@ def _exec(cmd: str, root: str | os.PathLike | None = None) -> list[str]:
 
 
 def get_branches(root: str | os.PathLike | None = None) -> list[str]:
+    """Return list of branch names in the git repository"""
     branches = _exec("git branch -l --format '%(refname:short)'", root=root)
     if branches:
         return branches
@@ -79,6 +84,7 @@ def get_branches(root: str | os.PathLike | None = None) -> list[str]:
 
 
 def get_branch(root: str | os.PathLike | None = None) -> str | None:
+    """Return branch name pointing to HEAD, or None"""
     branches = _exec("git rev-parse --abbrev-ref HEAD", root=root)
     if branches:
         return branches[0]
@@ -86,25 +92,43 @@ def get_branch(root: str | os.PathLike | None = None) -> str | None:
 
 
 def get_all_tags(sort_by: str = DEFAULT_SORT_BY, root: str | os.PathLike | None = None) -> list[str]:
+    """Return list of tags in the git repository"""
     tags = _exec(f"git tag --sort=-{sort_by}", root=root)
     if tags:
         return tags
     return []
 
 
-@deprecated(version="1.8.0", reason="will be dropped in v2.0.0. Please use 'get_tags' instead")
 def get_branch_tags(*args, **kwargs) -> list[str]:
+    warnings.warn(
+        "`get_branch_tags` function is deprecated "
+        "since setuptools-git-versioning v1.8.0 "
+        "and will be dropped in v2.0.0\n"
+        "Please use `get_tags` instead",
+        category=UserWarning,
+    )
+
     return get_tags(*args, **kwargs)
 
 
-def get_tags(sort_by: str = DEFAULT_SORT_BY, root: str | os.PathLike | None = None) -> list[str]:
+def get_tags(
+    sort_by: str = DEFAULT_SORT_BY,
+    filter_callback: Callable[[str], str | None] | None = None,
+    root: str | os.PathLike | None = None,
+) -> list[str]:
+    """Return list of tags merged into HEAD history tree"""
     tags = _exec(f"git tag --sort=-{sort_by} --merged", root=root)
+    if filter_callback:
+        # pull the tags that don't start with tag_prefix out of the list
+        mytags = list(filter(filter_callback, tags))
+        return mytags
     if tags:
         return tags
     return []
 
 
 def get_tag(*args, **kwargs) -> str | None:
+    """Return latest tag merged into HEAD history tree"""
     tags = get_tags(*args, **kwargs)
     if tags:
         return tags[0]
@@ -112,6 +136,7 @@ def get_tag(*args, **kwargs) -> str | None:
 
 
 def get_sha(name: str = "HEAD", root: str | os.PathLike | None = None) -> str | None:
+    """Get commit SHA-1 hash"""
     sha = _exec(f'git rev-list -n 1 "{name}"', root=root)
     if sha:
         return sha[0]
@@ -119,6 +144,7 @@ def get_sha(name: str = "HEAD", root: str | os.PathLike | None = None) -> str |
 
 
 def get_latest_file_commit(path: str | os.PathLike, root: str | os.PathLike | None = None) -> str | None:
+    """Get SHA-1 hash of latest commit of the file in the repository"""
     sha = _exec(f'git log -n 1 --pretty=format:%H -- "{path}"', root=root)
     if sha:
         return sha[0]
@@ -126,6 +152,7 @@ def get_latest_file_commit(path: str | os.PathLike, root: str | os.PathLike | No
 
 
 def is_dirty(root: str | os.PathLike | None = None) -> bool:
+    """Check index status, and return True if there are some uncommitted changes"""
     res = _exec("git status --short", root=root)
     if res:
         return True
@@ -133,18 +160,19 @@ def is_dirty(root: str | os.PathLike | None = None) -> bool:
 
 
 def count_since(name: str, root: str | os.PathLike | None = None) -> int | None:
+    """Get number of commits between HEAD and the commit, or None if they are not related"""
     res = _exec(f'git rev-list --count HEAD "^{name}"', root=root)
     if res:
         return int(res[0])
     return None
 
 
-def set_default_options(config: dict):
+def _set_default_options(config: dict):
     for key, value in DEFAULT_CONFIG.items():
         config.setdefault(key, value)
 
 
-def read_toml(name_or_path: str | os.PathLike = "pyproject.toml", root: str | os.PathLike | None = None) -> dict:
+def _read_toml(name_or_path: str | os.PathLike = "pyproject.toml", root: str | os.PathLike | None = None) -> dict:
     file_path = Path(root or os.getcwd()).joinpath(name_or_path)
     if not file_path.exists():
         log.log(INFO, "'%s' does not exist", file_path)
@@ -154,16 +182,24 @@ def read_toml(name_or_path: str | os.PathLike = "pyproject.toml", root: str | os
         raise OSError(f"'{file_path}' is not a file")
 
     log.log(INFO, "Trying 'pyproject.toml' ...")
-    import toml
+    try:
+        # for Python 3.11
+        import tomllib
+
+        with file_path.open("rb") as file:
+            parsed_file = tomllib.load(file)
+    except (ImportError, NameError):
+        import toml  # type: ignore[no-redef]
+
+        parsed_file = toml.load(file_path)
 
-    parsed_file = toml.load(file_path)
     result = parsed_file.get("tool", {}).get("setuptools-git-versioning", None)
     if result:
         log.log(DEBUG, "'tool.setuptools-git-versioning' section content:\n%s", pformat(result))
     return result
 
 
-def infer_setup_py(name_or_path: str = "setup.py", root: str | os.PathLike | None = None) -> str | None:
+def _infer_setup_py(name_or_path: str = "setup.py", root: str | os.PathLike | None = None) -> str | None:
     root_path = Path(root or os.getcwd())
     file_path = root_path.joinpath(name_or_path)
     if not file_path.exists():
@@ -181,7 +217,7 @@ def infer_setup_py(name_or_path: str = "setup.py", root: str | os.PathLike | Non
     original_sys_path = sys.path.copy()
     original_sys_modules = sys.modules.copy()
     try:
-        add_to_sys_path(root_path)
+        _add_to_sys_path(root_path)
         os.chdir(root_path)
         dist = run_setup(os.fspath(file_path), stop_after="init")
         return infer_version(dist, root=root)
@@ -192,7 +228,7 @@ def infer_setup_py(name_or_path: str = "setup.py", root: str | os.PathLike | Non
 
 
 # TODO: remove along with version_config
-def parse_config(dist: Distribution, attr: Any, value: Any) -> None:
+def _parse_config(dist: Distribution, attr: Any, value: Any) -> None:
     from distutils.errors import DistutilsOptionError
 
     if attr == "version_config" and value is not None:
@@ -211,7 +247,6 @@ def parse_config(dist: Distribution, attr: Any, value: Any) -> None:
             )
 
 
-# real version is generated here
 def infer_version(dist: Distribution, root: str | os.PathLike | None = None) -> str | None:
     log.log(INFO, "Trying 'setup.py' ...")
 
@@ -229,7 +264,7 @@ def infer_version(dist: Distribution, root: str | os.PathLike | None = None) ->
         )
         config = {"enabled": config}
 
-    toml_config = read_toml(root=root)
+    toml_config = _read_toml(root=root)
 
     if config is None:
         config = toml_config
@@ -250,61 +285,61 @@ def infer_version(dist: Distribution, root: str | os.PathLike | None = None) ->
         # Nothing to do here
         return None
 
-    set_default_options(config)
+    _set_default_options(config)
 
     version = version_from_git(dist.metadata.name, **config, root=root)
     dist.metadata.version = version
     return version
 
 
-def read_version_from_file(name_or_path: str | os.PathLike, root: str | os.PathLike | None = None) -> str:
+def _read_version_from_file(name_or_path: str | os.PathLike, root: str | os.PathLike | None = None) -> str:
     return Path(root or os.getcwd()).joinpath(name_or_path).read_text().strip()
 
 
-def substitute_env_variables(template: str) -> str:
-    log.log(DEBUG, "Substitute environment variables in template '%s'", template)
+def _substitute_env_variables(template: str) -> str:
+    log.log(DEBUG, "Substitute environment variables in template %r", template)
     for var, default in ENV_VARS_REGEXP.findall(template):
-        log.log(DEBUG, "Variable: '%s'", var)
+        log.log(DEBUG, "Variable: %r", var)
 
         if default.upper() == "IGNORE":
             default = ""
         elif not default:
             default = "UNKNOWN"
-        log.log(DEBUG, "Default: '%s'", default)
+        log.log(DEBUG, "Default: %r", default)
 
         value = os.environ.get(var, default)
-        log.log(DEBUG, "Value: '%s'", value)
+        log.log(DEBUG, "Value: %r", value)
 
         template, _ = ENV_VARS_REGEXP.subn(value, template, count=1)
 
-    log.log(DEBUG, "Result: '%s'", template)
+    log.log(DEBUG, "Result: %r", template)
     return template
 
 
-def substitute_timestamp(template: str) -> str:
-    log.log(DEBUG, "Substitute timestampts in template '%s'", template)
+def _substitute_timestamp(template: str) -> str:
+    log.log(DEBUG, "Substitute timestamps in template %r", template)
 
     now = datetime.now()
     for fmt in TIMESTAMP_REGEXP.findall(template):
         format_string = fmt or "%s"
-        log.log(DEBUG, "Format: '%s'", format_string)
+        log.log(DEBUG, "Format: %r", format_string)
 
-        result = now.strftime(fmt or "%s")
-        log.log(DEBUG, "Value: '%s'", result)
+        result = now.strftime(format_string)
+        log.log(DEBUG, "Value: %r", result)
 
         template, _ = TIMESTAMP_REGEXP.subn(result, template, count=1)
 
-    log.log(DEBUG, "Result: '%s'", template)
+    log.log(DEBUG, "Result: %r", template)
     return template
 
 
-def resolve_substitutions(template: str, *args, **kwargs) -> str:
-    log.log(DEBUG, "Template: '%s'", template)
+def _resolve_substitutions(template: str, *args, **kwargs) -> str:
+    log.log(DEBUG, "Template: %r", template)
     log.log(DEBUG, "Args:%s", pformat(args))
 
     while True:
         if "{env" in template:
-            new_template = substitute_env_variables(template)
+            new_template = _substitute_env_variables(template)
             if new_template == template:
                 break
             else:
@@ -313,19 +348,19 @@ def resolve_substitutions(template: str, *args, **kwargs) -> str:
             break
 
     if "{timestamp" in template:
-        template = substitute_timestamp(template)
+        template = _substitute_timestamp(template)
 
     return template.format(*args, **kwargs)
 
 
-def add_to_sys_path(root: str | os.PathLike | None) -> None:
+def _add_to_sys_path(root: str | os.PathLike | None) -> None:
     root_path = os.fspath(Path(root or os.getcwd()))
     if root_path not in sys.path:
         log.log(DEBUG, "Adding '%s' folder to sys.path", root_path)
         sys.path.insert(0, root_path)
 
 
-def import_reference(
+def _import_reference(
     ref: str,
     package_name: str | None = None,
     root: str | os.PathLike | None = None,
@@ -333,7 +368,7 @@ def import_reference(
     if ":" not in ref:
         raise NameError(f"Wrong reference name: {ref}")
 
-    add_to_sys_path(root)
+    _add_to_sys_path(root)
 
     module_name, attr = ref.split(":")
     log.log(DEBUG, "Executing 'from %s.%s import %s'", package_name or "", module_name, attr)
@@ -342,115 +377,132 @@ def import_reference(
     return getattr(module, attr)
 
 
-def load_callable(
+def _load_callable(
     inp: str,
     package_name: str | None = None,
     root: str | os.PathLike | None = None,
 ) -> Callable:
-    ref = import_reference(inp, package_name, root=root)
+    ref = _import_reference(inp, package_name, root=root)
     if not callable(ref):
         raise ValueError(f"{ref} of type {type(ref)} is not callable")
 
     return ref
 
 
-def load_tag_formatter(
-    tag_formatter: str | Callable[[str], str],
+def _branch_formatter_factory(regexp: str) -> Callable[[str], str]:
+    pattern = re.compile(regexp)
+
+    def branch_formatter(branch: str) -> str:
+        match = pattern.match(branch)
+        if match:
+            return match.group("branch")
+
+        raise ValueError(f"Branch name {branch} does not match regexp '{regexp}'")
+
+    return branch_formatter
+
+
+def _tag_formatter_factory(regexp: str) -> Callable[[str], str]:
+    pattern = re.compile(regexp)
+
+    def tag_formatter(tag: str) -> str:
+        match = pattern.match(tag)
+        if match:
+            return match.group("tag")
+
+        raise ValueError(f"Tag name {tag} does not match regexp '{regexp}'")
+
+    return tag_formatter
+
+
+def _tag_filter_factory(regexp: str) -> Callable[[str], str | None]:
+    pattern = re.compile(regexp)
+
+    def tag_filter(tag: str) -> str | None:
+        match = pattern.match(tag)
+        if match:
+            log.info("Matched %s", tag)
+            return tag
+        else:
+            return None
+
+    return tag_filter
+
+
+def _callable_factory(
+    callable_name: str,
+    regexp_or_ref: str | Callable,
+    callable_factory: Callable[[str], Callable],
     package_name: str | None = None,
     root: str | os.PathLike | None = None,
 ) -> Callable:
-    log.log(INFO, "Parsing tag_formatter '%s' of type '%s'", tag_formatter, type(tag_formatter).__name__)
+    log.log(INFO, "Parsing %s %r of type %r", callable_name, regexp_or_ref, type(regexp_or_ref).__name__)
 
-    if callable(tag_formatter):
-        log.log(DEBUG, "Value is callable with signature %s", inspect.Signature.from_callable(tag_formatter))
-        return tag_formatter
+    if callable(regexp_or_ref):
+        log.log(DEBUG, "Value is callable with signature %s", inspect.Signature.from_callable(regexp_or_ref))
+        return regexp_or_ref
 
     try:
-        return load_callable(tag_formatter, package_name, root=root)
+        return _load_callable(regexp_or_ref, package_name, root=root)
     except (ImportError, NameError) as e:
-        log.warning("tag_formatter is not a valid function reference: %s", e)
+        log.warning("%s is not a valid function reference: %s", callable_name, e)
 
     try:
-        pattern = re.compile(tag_formatter)
-
-        def formatter(tag):
-            match = pattern.match(tag)
-            if match:
-                return match.group("tag")
-
-            raise ValueError(f"Tag name {tag} does not match regexp '{tag_formatter}'")
-
-        return formatter
+        return callable_factory(regexp_or_ref)
     except re.error as e:
-        log.error("tag_formatter is not valid regexp: %s", e)
-
-    raise ValueError("Cannot parse tag_formatter")
-
-
-def load_branch_formatter(
-    branch_formatter: str | Callable[[str], str],
-    package_name: str | None = None,
-    root: str | os.PathLike | None = None,
-) -> Callable:
-    log.log(INFO, "Parsing branch_formatter '%s' of type '%s'", branch_formatter, type(branch_formatter).__name__)
-
-    if callable(branch_formatter):
-        log.log(DEBUG, "Value is callable with signature %s", inspect.Signature.from_callable(branch_formatter))
-        return branch_formatter
-
-    try:
-        return load_callable(branch_formatter, package_name, root=root)
-    except (ImportError, NameError) as e:
-        log.warning("branch_formatter is not a valid function reference: %s", e)
-
-    try:
-        pattern = re.compile(branch_formatter)
-
-        def formatter(branch):
-            match = pattern.match(branch)
-            if match:
-                return match.group("branch")
-
-            raise ValueError(f"Branch name {branch} does not match regexp '{branch_formatter}'")
-
-        return formatter
-    except re.error as e:
-        log.error("branch_formatter is not valid regexp: %s", e)
-
-    raise ValueError("Cannot parse branch_formatter")
+        log.error("%s is not valid regexp: %s", callable_name, e)
+        raise ValueError(f"Cannot parse {callable_name}") from e
 
 
 # TODO: return Version object instead of str
-def get_version_from_callback(
+def _get_version_from_callback(
     version_callback: str | Callable[[], str],
     package_name: str | None = None,
     root: str | os.PathLike | None = None,
 ) -> str:
-    log.log(INFO, "Parsing version_callback %s of type %s", version_callback, type(version_callback))
+    log.log(INFO, "Parsing version_callback %r of type %r", version_callback, type(version_callback).__name__)
 
     if callable(version_callback):
         log.log(DEBUG, "Value is callable with signature %s", inspect.Signature.from_callable(version_callback))
         result = version_callback()
     else:
-
         log.log(INFO, "Is not callable, trying to import ...")
         result = version_callback
 
         try:
-            callback = load_callable(version_callback, package_name, root=root)
+            callback = _load_callable(version_callback, package_name, root=root)
             result = callback()
         except ValueError as e:
             log.log(INFO, "Is not a callable")
             log.log(DEBUG, str(e))
             log.log(INFO, "Assuming it is a string attribute")
-            result = import_reference(version_callback, package_name, root=root)
+            result = _import_reference(version_callback, package_name, root=root)
         except (ImportError, NameError) as e:
             log.warning("version_callback is not a valid reference: %s", e)
 
+    return _sanitize_version(result)
+
+
+def _sanitize_version(version: str) -> str:
     from packaging.version import Version
 
-    log.log(INFO, "Result %s", result)
-    return Version(result).public
+    log.log(INFO, "Before sanitization %r", version)
+
+    public, sep, local = version.partition("+")
+
+    # replace "feature/ABC-123" with "feature.ABC.123"
+    sanitized_public = UNSUPPORTED_SYMBOL_REGEXP.sub(".", public)
+    sanitized_local = UNSUPPORTED_SYMBOL_REGEXP.sub(".", local)
+
+    sanitized_version = sanitized_public + sep + sanitized_local
+    sanitized_version = sanitized_version.rstrip(".")
+
+    # replace "feature.ABC.123" with "feature.abc.123"
+    # drop leading "v" symbol
+    # other replacements according to PEP-440, like "-dev" -> ".dev"
+    result = str(Version(sanitized_version))
+    log.log(INFO, "Result %r", result)
+    return result
 
 
 # TODO: return Version object instead of str
@@ -463,8 +515,9 @@ def version_from_git(
     version_callback: str | Callable[[], str] | None = None,
     version_file: str | os.PathLike | None = None,
     count_commits_from_version_file: bool = False,
-    tag_formatter: Callable[[str], str] | None = None,
-    branch_formatter: Callable[[str], str] | None = None,
+    tag_formatter: Callable[[str], str] | str | None = None,
+    branch_formatter: Callable[[str], str] | str | None = None,
+    tag_filter: Callable[[str], str | None] | str | None = None,
     sort_by: str = DEFAULT_SORT_BY,
     root: str | os.PathLike | None = None,
 ) -> str:
@@ -476,7 +529,8 @@ def version_from_git(
         for line in lines:
             if line.startswith("Version:"):
                 result = line[8:].strip()
-                log.log(INFO, "Return '%s'", result)
+                log.log(INFO, "Return %r", result)
+                # running on sdist package, do not sanitize
                 return result
 
     if version_callback is not None:
@@ -484,74 +538,96 @@ def version_from_git(
             raise ValueError(
                 "Either 'version_file' or 'version_callback' can be passed, but not both at the same time",
             )
-        return get_version_from_callback(version_callback, package_name, root=root)
+        return _get_version_from_callback(version_callback, package_name, root=root)
+
+    filter_callback = None
+    if tag_filter:
+        filter_callback = _callable_factory(
+            callable_name="tag_filter",
+            regexp_or_ref=tag_filter,
+            callable_factory=_tag_filter_factory,
+            package_name=package_name,
+            root=root,
+        )
 
     from_file = False
     log.log(INFO, "Getting latest tag")
-    log.log(DEBUG, "Sorting tags by '%s'", sort_by)
-    tag = get_tag(sort_by=sort_by, root=root)
+    log.log(DEBUG, "Sorting tags by %r", sort_by)
+    tag = get_tag(sort_by=sort_by, root=root, filter_callback=filter_callback)
 
     if tag is None:
         log.log(INFO, "No tag, checking for 'version_file'")
         if version_file is None:
-            log.log(INFO, "No 'version_file' set, return starting_version '%s'", starting_version)
-            return starting_version
+            log.log(INFO, "No 'version_file' set, return starting_version %r", starting_version)
+            return _sanitize_version(starting_version)
 
         if not Path(version_file).exists():
             log.log(
                 INFO,
-                "version_file '%s' does not exist, return starting_version '%s'",
+                "version_file '%s' does not exist, return starting_version %r",
                 version_file,
                 starting_version,
             )
-            return starting_version
+            return _sanitize_version(starting_version)
 
         log.log(INFO, "version_file '%s' does exist, reading its content", version_file)
         from_file = True
-        tag = read_version_from_file(version_file, root=root)
+        tag = _read_version_from_file(version_file, root=root)
 
         if not tag:
-            log.log(INFO, "File is empty, return starting_version '%s'", version_file, starting_version)
-            return starting_version
+            log.log(INFO, "File is empty, return starting_version %r", version_file, starting_version)
+            return _sanitize_version(starting_version)
 
-        log.log(DEBUG, "File content: '%s'", tag)
+        log.log(DEBUG, "File content: %r", tag)
         if not count_commits_from_version_file:
-            result = VERSION_PREFIX_REGEXP.sub("", tag)  # for tag "v1.0.0" drop leading "v" symbol
-            log.log(INFO, "Return '%s'", result)
-            return result
+            return _sanitize_version(tag)
 
         tag_sha = get_latest_file_commit(version_file, root=root)
-        log.log(DEBUG, "File content: '%s'", tag)
+        log.log(DEBUG, "File SHA-256: %r", tag_sha)
     else:
-        log.log(INFO, "Latest tag: '%s'", tag)
+        log.log(INFO, "Latest tag: %r", tag)
         tag_sha = get_sha(tag, root=root)
-        log.log(INFO, "Tag SHA-256: '%s'", tag_sha)
+        log.log(INFO, "Tag SHA-256: %r", tag_sha)
 
         if tag_formatter is not None:
-            tag_fmt = load_tag_formatter(tag_formatter, package_name, root=root)
-            tag = tag_fmt(tag)
-            log.log(DEBUG, "Tag after formatting: '%s'", tag)
+            tag_format_callback = _callable_factory(
+                callable_name="tag_formatter",
+                regexp_or_ref=tag_formatter,
+                callable_factory=_tag_formatter_factory,
+                package_name=package_name,
+                root=root,
+            )
+
+            tag = tag_format_callback(tag)
+            log.log(DEBUG, "Tag after formatting: %r", tag)
 
     dirty = is_dirty(root=root)
-    log.log(INFO, "Is dirty: %s", dirty)
+    log.log(INFO, "Is dirty: %r", dirty)
 
     head_sha = get_sha(root=root)
-    log.log(INFO, "HEAD SHA-256: '%s'", head_sha)
+    log.log(INFO, "HEAD SHA-256: %r", head_sha)
 
     full_sha = head_sha if head_sha is not None else ""
     ccount = count_since(tag_sha, root=root) if tag_sha is not None else None
-    log.log(INFO, "Commits count between HEAD and latest tag: %s", ccount)
+    log.log(INFO, "Commits count between HEAD and latest tag: %r", ccount)
 
     on_tag = head_sha is not None and head_sha == tag_sha and not from_file
-    log.log(INFO, "HEAD is tagged: %s", on_tag)
+    log.log(INFO, "HEAD is tagged: %r", on_tag)
 
     branch = get_branch(root=root)
-    log.log(INFO, "Current branch: '%s'", branch)
+    log.log(INFO, "Current branch: %r", branch)
 
     if branch_formatter is not None and branch is not None:
-        branch_fmt = load_branch_formatter(branch_formatter, package_name, root=root)
-        branch = branch_fmt(branch)
-        log.log(INFO, "Branch after formatting: '%s'", branch)
+        branch_format_callback = _callable_factory(
+            callable_name="branch_formatter",
+            regexp_or_ref=branch_formatter,
+            callable_factory=_branch_formatter_factory,
+            package_name=package_name,
+            root=root,
+        )
+
+        branch = branch_format_callback(branch)
+        log.log(INFO, "Branch after formatting: %r", branch)
 
     if dirty:
         log.log(INFO, "Using template from 'dirty_template' option")
@@ -563,35 +639,22 @@ def version_from_git(
         log.log(INFO, "Using template from 'template' option")
         t = template
 
-    version = resolve_substitutions(t, sha=full_sha[:8], tag=tag, ccount=ccount, branch=branch, full_sha=full_sha)
-    log.log(INFO, "Version number after resolving substitutions: '%s'", version)
-
-    # Ensure local version label only contains permitted characters
-    public, sep, local = version.partition("+")
-    local_sanitized = LOCAL_REGEXP.sub(".", local)
-    if local_sanitized != local:
-        log.log(INFO, "Local version part after sanitization: '%s'", local_sanitized)
-
-    public_sanitized = VERSION_PREFIX_REGEXP.sub("", public)  # for version "v1.0.0" drop leading "v" symbol
-    if public_sanitized != public:
-        log.log(INFO, "Public version part after sanitization: '%s'", public_sanitized)
-
-    result = (public_sanitized + sep + local_sanitized) or "0.0.0"
-    log.log(INFO, "Result: '%s'", result)
-    return result
+    version = _resolve_substitutions(t, sha=full_sha[:8], tag=tag, ccount=ccount, branch=branch, full_sha=full_sha)
+    log.log(INFO, "Version number after resolving substitutions: %r", version)
+    return _sanitize_version(version)
 
 
-def main(config: dict | None = None, root: str | os.PathLike | None = None) -> Version:
+def get_version(config: dict | None = None, root: str | os.PathLike | None = None) -> Version:
     from packaging.version import Version
 
     if not config:
         log.log(INFO, "No explicit config passed")
         log.log(INFO, "Searching for config files in '%s' folder", root or os.getcwd())
-        result = infer_setup_py(root=root)
+        result = _infer_setup_py(root=root)
         if result is not None:
             return Version(result)
 
-        config = read_toml(root=root)
+        config = _read_toml(root=root)
 
     if not config or not config.pop("enabled", True):
         raise RuntimeError(
@@ -608,7 +671,7 @@ def main(config: dict | None = None, root: str | os.PathLike | None = None) -> V
     if not isinstance(config, dict):
         raise RuntimeError(f"Wrong config format. Expected dict, got: {config}")
 
-    set_default_options(config)
+    _set_default_options(config)
     result = version_from_git(**config, root=root)
     return Version(result)
 
@@ -640,7 +703,7 @@ def __main__():
     namespace = parser.parse_args()
     log_level = VERBOSITY_LEVELS.get(namespace.verbose, logging.DEBUG)
     logging.basicConfig(level=log_level, format=LOG_FORMAT, stream=sys.stderr)
-    print(main(root=namespace.root).public)
+    print(str(get_version(root=namespace.root)))
 
 
 if __name__ == "__main__":
